/*  BOUND2.PRG 

Gauss Code for "Shrinkage Efficiency Bounds" 
by Bruce E. Hansen
Department of Economics, University of Wisconsin

This program calculates Table 1, part of Table 2.
The results are also used to calculate the Figures 
*/

library pgraph;
pqgwin many;
graphset;
_pdate="";
_plwidth=8;
let _pmcolor = 0 0 0 0 0 0 0 0 15;
fonts("complex simgrma");
_pcolor=1;

declare _p,_psi;

qn=20000;
pn=10;
psimax=20;
gd=psimax*100;

aLK=zeros(pn,1);
aGP=zeros(pn,2);
aS=zeros(pn,1);
thetaM=zeros(pn,3);
thetaM2=zeros(pn,2);
nuA=zeros(pn,1);
rKT=zeros(pn,1);
mregret=zeros(11,pn);

for i (3,pn,1);
  p=i;
  psimaxp=psimax*p;
  psi=seqa(0,psimaxp/gd,gd);
  r2=zeros(gd,1);
  r2[1]=Rbound(p,0,qn);
  for gi (2,gd,1);
    psig=psi[gi];
    r2[gi]=Rbound(p,psig,qn);
  endfor;
  regret=zeros(gd,12);

  q2=200; cd=0;
  do until cd .>= .9999;
    q2=q2+100;
    cd=cdfchinc(q2,p+2,sqrt(psimaxp));
  endo; 
  inc=q2/(qn-1);
  q=seqa(inc,inc,qn);
  f1=zeros(qn,gd);
  f2=zeros(qn,gd);
  f1[.,1]=pdfchic(q,p)*inc;
  for gi (2,gd,1);
    psig=psi[gi];
    f1[.,gi]=pdfchinc(q,p,psig)*inc;
    f2[.,gi]=pdfchinc(q,p+2,psig)*(psig*2*inc);
  endfor;

@ James-Stein (1961) @
  rjs=Risk_JS0(p,psi)-r2;
  mregret[1,i]=maxc(rjs);
  regret[.,1]=rjs;

@ Kubokawa (1991) @
  phi=phi_S(p,q,2);
  h=phi./q;
  rk=p+f1'((h-2).*phi)+f2'h-r2;
  mregret[2,i]=maxc(rk);
  regret[.,2]=rk;

@ Li-Kuo (1982) @
  rn=(p/2-1)*100;
  a1=seqa(0,.01,rn)';
  c1=a1.*(2.^(a1/2)).*gamma(p/2-1-a1/2)./gamma(p/2-1-a1);
  m1=MomentChiNC(-1-a1/2,p,psi);
  m2=MomentChiNC(-1-a1,p,psi);
  rLK=rjs-2*c1.*a1.*m1+(c1.^2).*m2; 
  mrLK=maxc(rLK);
  ai=minindc(mrLK);
  aLK[i]=a1[ai];
  mregret[3,i]=mrLK[ai];
  regret[.,3]=rLK[.,ai];

@ Guo-Pal (1992); @
  mrGP=zeros(rn,rn)+mrLK;
  for a2i (2,rn,1);
    a2=a1[a2i];
    c2=(1-c1.*(gamma(p/2-1-(a1+a2)/2)./gamma(p/2-1-a2/2)
       ./a2./(2.^(a1/2)))).*c1[a2i];
    m3=MomentChiNC(-1-a1/2-a2/2,p,psi);
    rGP=rLK+(c2.^2).*m2[.,a2i]+2.*c1.*c2.*m3-2.*c2.*a2.*m1[.,a2i];
    rg=maxc(rGP);
    mrGP[1:a2i-1,a2i]=rg[1:a2i-1];
  endfor;   
  a2i=minindc(minc(mrGP));
  a1i=minindc(minc(mrGP'));
  aa1=a1[a1i];
  aa2=a1[a2i];
  aGP[i,1]=aa1;
  aGP[i,2]=aa2;
  mregret[4,i]=mrGP[a1i,a2i];

@ Alam (1973) @
  nu1=(2*p+5-sqrt(4*p*p+8*p-7))/4;
  nu1=round(nu1*100)/100;
  rn=(1-nu1)*100;
  nu=seqa(nu1,.01,rn);
  rA=zeros(gd,rn);
  for ni (1,rn,1);
    h=h_A(p,q,nu[ni]);
    rA[.,ni]=p+f1'((h-2).*(q.*h))+f2'h-r2;
  endfor;
  mrA=maxc(rA);
  ni=minindc(mrA);
  nuA[i]=nu[ni];
  mregret[5,i]=mrA[ni];
  regret[.,5]=rA[.,ni];

@ Kuriki-Takemura (2000) @
  rn=300;
  r=seqa(.01,.01,rn);
  phi=phi_KT(p,q,r');
  h=phi./q;
  rgKT=p+f1'((h-2).*phi)+f2'h-r2;
  mrKT=maxc(rgKT);
  ri=minindc(mrKT);
  rKT[i]=r[ri];
  mregret[6,i]=mrKT[ri];
  regret[.,6]=rgKT[.,ri];

@ Baranchik (1964) @
  rb=Risk_JS(p,psi,p-2)-r2;
  mregret[7,i]=maxc(rb);
  regret[.,7]=rb;

@ Shao and Strawderman @
  rs=Risk_SS(p,psi)-r2;
  mregret[8,i]=maxc(rs);
  regret[.,8]=rs;

@ Strawderman (1971), Marayuma (1998) and Marayuma (2004) @
  eps=1e-8;
  n1=maxc((3-p/2)|0);
  na=1+(2-n1)*100;
  aa=seqa(n1,.01,na);
  nb=20;
  bb=seqa(0,-.01,nb);
  mrM=zeros(na,nb);
  mbrM=zeros(na,nb);
  betastore=zeros(na,nb);
  for ai (1,na,1);
    a=aa[ai];
    for bi(1,nb,1);
      b=bb[bi];
      d=(b*(3*p/2+1-a)+(a+p/2-3))/(b+1);
      if d<(-eps);
        mrM[ai,bi]=maxc(p-r2);
        mbrM[ai,bi]=maxc(p-r2);
      else;
        dd=1+d*(p/2-a+1)*(b+1)/(b-a+p/2+2)/(b-a+p/2+3);
        betastar=(sqrt(dd)-1)*gamma(p/2-a+1)*gamma(b+1)/gamma(b-a+p/2+2);
        inc=.001;
        betan=round(betastar/inc)+1;
        betastar=(betan-1)*inc;
        beta=seqa(0,inc,betan);
        h=h_M(p,q,a,b,beta');
        rk=p+f1'((h-2).*(q.*h))+f2'h-r2;
        rg=maxc(rk);
        mbrM[ai,bi]=rg[1];
        mrM[ai,bi]=minc(rg);
        betastore[ai,bi]=beta[minindc(rg)];
      endif;
    endfor;
  endfor;

@ Strawderman (1971) @
  ai=minindc(mbrM[.,1]);
  a=aa[ai];
  aS[i]=a;
  mregret[9,i]=mbrM[ai,1];
  h=h_M(p,q,a,0,0);
  regret[.,9]=p+f1'((h-2).*(q.*h))+f2'h-r2;

@ Marayuma (1998) @
  bi2=minindc(minc(mbrM));
  ai2=minindc(minc(mbrM'));
  a=aa[ai2];
  b=bb[bi2];
  thetaM2[i,1]=a;
  thetaM2[i,2]=b;
  mregret[10,i]=mbrM[ai2,bi2];
  h=h_M(p,q,a,b,0);
  regret[.,10]=p+f1'((h-2).*(q.*h))+f2'h-r2;

@ Marayuma (2004) @
  bi=minindc(minc(mrM));
  ai=minindc(minc(mrM'));
  a=aa[ai];
  b=bb[bi];
  beta=betastore[ai,bi];
  thetaM[i,1]=a;
  thetaM[i,2]=b;
  thetaM[i,3]=beta;
  mregret[11,i]=mrM[ai,bi];
  h=h_M(p,q,a,b,beta);
  regret[.,11]=p+f1'((h-2).*(q.*h))+f2'h-r2;

tit="" $+ "Regret, p = " $+ ftocv(p,1,0);
title(tit);
xy(psi,regret);

/***************************************************************/

output file=bound2.out reset;
"";
format /mb1 /rds 6,0;
"p:";
seqa(3,1,i-2)';
"";
format /mb1 /rds 6,2;
"Li-Kuo (1982) a1";
aLK[3:i]';"";
"Guo-Pal (1992) a1, a2";
aGP[3:i,.]';"";
"Alam (1973) nu";
nuA[3:i]';"";
"Kuriki-Takemura (2000) r";
rKT[3:i]';"";
"Strawderman (1971) a";
aS[3:i]';"";
"Maruyama (1998) a,b";
thetaM2[3:i,.]';"";
"Maruyama (2004) a,b,beta ";
thetaM[3:i,1 2]';
format /mb1 /rds 6,3;
thetaM[3:i,3]';
"";"";"";

format /mb1 /rds 6,3;
seqa(3,1,i-2)';
mregret[.,3:i];
output off;

save aLK,aS,thetaM,thetaM2,nuA,rKT,aGP;

endfor;


/************************************************************/

fn Risk_JS0(p,psi)=p-((p-2)^2).*MomentChiNC(-1,p,psi);

proc Risk_JS(p,psi,c);
local m0,m1,m2,r;
  m1=TrimmedMomentChiNC(1,0,c,p,psi);
  m0=TrimmedMomentChiNC(0,0,c,p,psi);
  m2=ITrimmedMomentChiNC(c,p,psi);
  r=p+m1-2*p.*m0+c.*(c-2*p+4).*m2;
retp(r);
endp;


proc Risk_SS(p,psi);
local ps,a,pss,knots,theta,r;
  ps=p-.3;
  if p==3;
    a=0.0341; 
  elseif p==4;
    a=0.0393;
  elseif p==5;
    a=0.0566;
  else;
    a=0.0337;
  endif;
  pss=2*(p-1)-ps;
  knots=(p-2)|pss|(p-1)|ps|p;
  theta=1|(-1-a)|(2*a)|(-2*a)|(2*a);
  r=Risk_2(theta,knots);
retp(r);
endp;

proc Risk_2(theta,knots);
local n,p2,a,b,c,r,knot1,m0,m1,ms,m_1;
  p2=2*(p-2);
  n=rows(knots);
  if n>1; knot1=0|knots[1:n-1];else;knot1=0;endif;
  a=-cumsumc(theta.*knot1);
  b=cumsumc(theta);
  c=theta'(maxc(knots)-knot1);
  m0 = TrimmedMomentChiNC(0,knot1',knots',p,psi);
  m1 = TrimmedMomentChiNC(1,knot1',knots',p,psi);
  ms = ITrimmedMomentChiNC(maxc(knots),p,psi);
  r=p+m0*(2*b.*(a-p))+m1*(b.^2)-ms*c*(p2-c);
  if n>1;
    m_1= TrimmedMomentChiNC(-1,knot1',knots',p,psi);
    r=r+m_1*(a.*(a-p2));
  endif;
retp(r);
endp;


/************************
Proc MomentChiNC(a,p,psi)
Proc TrimmedMomentChiNC(a,c1,c2,p,psi)
Proc ITrimmedMomentChiNC(a,c,p,psi)

Computes moments, trimmed moments, and CDF of the non-central chi-square distribution

For Q distributed as non-central chi-square with p degrees of freedom
and non-centrality parameter psi

MomentChiNC(a,p,psi)=E[(Q.^a)]
TrimmedMomentChiNC(a,c1,c2,p,psi)=E[(Q.^a)1(c1<=Q<=c2)]
ITrimmedMomentChiNC(c,p,psi)=E[(Q.^-1)1(c<=Q)]

Inputs:
a	moment
c,c1,c2	threshold
p	degrees of freedom
psi	non-centrality parameter

Inputs can be vector or matrix-valued, if element-by-element conformable

c1=0 is okay in TrimmedMomentChiNC(a,c1,c2,p,psi)
Note
TrimmedMomentChiNC(0,0,c,p,psi)=P(Q<=c)

The moments are computed using a truncated series expansion.
The series is computed up to 1000 terms, or until the additional
term is less than a tolerance, set at 1e-8. 

*******************/

proc MomentChiNC(a,p,psi);
local jn,tol,r,f,j,m,f0,r0;
  jn=1000;
  tol=1e-8;
  r0=exp(-psi/2);
  f0=r0;
  r=(2.^a).*gamma(a+p/2)./gamma(p/2).*r0;
  f=r;
  for j (1,jn,1);
    r=r.*psi.*((a+p/2+j-1)./(p/2+j-1)./j./2);
    f=f+r;
    r0=r0.*psi./(j*2);
    f0=f0+r0;
    m=maxc(vec(abs(f0-1)));
    if m<=tol; break;endif;
  endfor;
retp(f);
endp;

proc TrimmedMomentChiNC(a,c1,c2,p,psi);
local jn,tol,r,f,j,d,m,g,cg1,cg2,r0,f0;
  jn=1000;
  tol=1e-8;
  r0=exp(-psi/2);
  f0=r0;
  r=(2.^a).*gamma(a+p/2)./gamma(p/2).*r0;
  cg1=cdfgam(a+p/2,c1/2);
  cg2=cdfgam(a+p/2,c2/2);
  f=r.*(cg2-cg1);
  for j (1,jn,1);
    r=r.*psi.*((a+p/2+j-1)./(p/2+j-1)./j./2);
    cg1=cdfgam(a+p/2+j,c1/2);
    cg2=cdfgam(a+p/2+j,c2/2);
    d=r.*(cg2-cg1);
    f=f+d;
    r0=r0.*psi./(j*2);
    f0=f0+r0;
    m=maxc(vec(abs(f0-1)));
    if m<=tol; break;endif;
  endfor;
retp(f);
endp;


proc ITrimmedMomentChiNC(c,p,psi);
local jn,tol,r,f,j,d,m,g,cg1,r0,f0;
  jn=1000;
  tol=1e-8;
  r0=exp(-psi/2);
  f0=r0;
  r=r0./(p-2);
  cg1=cdfgam(p/2-1,c/2);
  f=r.*(1-cg1);
  for j (1,jn,1);
    r=r.*psi.*((p/2+j-2)./(p/2+j-1)./j./2);
    cg1=cdfgam(p/2+j-1,c/2);
    d=r.*(1-cg1);
    f=f+d;
    r0=r0.*psi./(j*2);
    f0=f0+r0;
    m=maxc(vec(abs(f0-1)));
    if m<=tol; break;endif;
  endfor;
retp(f);
endp;


/***********
Proc igam(a,x)

igam(a,x)=gamma(a,x)*exp(x)/x^a

where gamma(a,x) is the incomplete gamma function

*******************/

proc igam(a,x);
local tol,d,s2,ii,s,r,j;
  ii=1000;
  tol=1e-12;
  r=1./a;
  s=r;
  for j (1,ii,1);
    r=r.*(x./(j+a));
    s=s+r;
    d=maxc(vec(abs(r)));
    if d<=tol; break; endif;
  endfor;
retp(s);
endp;


proc ConfluentHypergeometricFunction(a,b,x);
local ii,s,r,j,n,ss,i,xi,aa,bb,c;
  ii=100;
  n=cols(x);
  r=rows(x);
  ss=zeros(r,n);
  for i (1,n,1);
    xi=x[.,i];
    aa=seqa(a,1,ii);
    bb=seqa(b,1,ii);
    j=seqa(1,1,ii);
    c=ones(1,r)|((aa./bb./j).*(xi'));
    ss[.,i]=sumc(cumprodc(c));
  endfor;
retp(ss);
endp;

proc h_M(p,q,a,b,beta);
local h,f1,f2,g1,g2,g3,g4,g5;
  g1=gamma(b+1);
  g2=gamma(p/2-a+2);
  g3=gamma(p/2-a+b+3);
  g4=gamma(p/2-a+1);
  g5=gamma(p/2-a+b+2);
  f1=ConfluentHypergeometricFunction(b+1,p/2-a+b+3,q/2)
    .*g1.*g2./g3;
  f2=ConfluentHypergeometricFunction(b+1,p/2-a+b+2,q/2)
    .*g1.*g4./g5;
  h=(f1+beta)./(f2+beta);
retp(h);
endp;

proc h_A(p,q,a);
local f1,f2,h;
  f1=ConfluentHypergeometricFunction(a,p/2+1,q/2);
  f2=ConfluentHypergeometricFunction(a,p/2,q/2);
  h=(f1./f2).*(1-2.*a./p);
retp(h);
endp;

fn phi_S(p,q,a)=(p-2*a+2)-2./igam((p/2-a+1),q/2);

proc phi_KT(p,q,r);
local phi,tau;
  phi=p-2-r./(sqrt(q)-r);
  tau=(((p-1)/(p-2))^2)*r;
  phi=phi.*(q.>tau);
retp(phi);
endp;

proc pdfchic(q,r); @ central chi-square density @
local k2,q2,f;
  k2=r/2;
  q2=q/2;
  f=exp(-q2).*(q2.^(k2-1))./(2*gamma(k2)); 
retp(f);
endp;

proc pdfchinc(q,r,psi); @ non-central chi-square density @
local q2,sd,sq,k2,n,a,i,f;
  k2=r/2;
  sq=sqrt(q);
  sd=sqrt(psi);
  n=floor(k2);
  a=k2-n;
  i=mbesselei(sq.*.sd,n,a); @ exponentially scaled Bessel function @
  i=reshape(i[.,n],rows(q),rows(psi));
  f=i.*exp(-((sq-sd').^2)/2).*((sq./sd').^(k2-1))/2; 
retp(f);
endp;

proc Rbound(p,psi,qn);
local q1,q2,ret1,ret2,inc,q,h,f,d,r;
    __output=0;
    __Tol=1e-12;
    _p=p;
    _psi=psi;
    if psi<=p;
      q1=0;ret1=1;
    else;
      q1=(psi-p)*p*(p+2)/(psi^2);
      {q1,ret1}=eqSolve(&hstar1,q1);
    endif;
    q2=2*(p-2)+psi/2-.5*sqrt((psi+4*(p-2))^2-16*((p-2)^2));
    {q2,ret2}=eqSolve(&hstar2,q2);
    inc=(q2-q1)/(qn-1);
    q=seqa(q1,inc,qn);
    h=hstar(q,p,psi);
    f=pdfchinc(q,p,psi); 
    d=sumc((h.^2).*f.*q)*inc;
    r=p-d-8*(p-2)*f[qn];
retp(r);
endp;


proc hstar(q,p,psi);
local h1,q0,sq,sd,n,i,p2;
if psi==0;
  h1=ones(rows(q),1);
else;
  q0=(q.==0);
  sq=sqrt(q+q0);
  sd=sqrt(psi);
  p2=p/2+1;
  n=floor(p2);
  i=mbesselei(sq.*sd,n,p2-n);
  h1=1-(i[.,n]./i[.,n-1]).*(sd./sq);
  h1=h1.*(1-q0)+q0*(1-psi/p);
endif;
retp(h1);
endp;

fn hstar1(q)=hstar(q,_p,_psi);
fn hstar2(q)=hstar(q,_p,_psi)-2*(_p-2)./q;

proc pdfchinc(q,r,psi);
local q2,sd,sq,k2,n,a,i,f;
k2=r/2;
if psi==0;
  q2=q/2;
  f=exp(-q2).*(q2.^(k2-1));
  f=f./(2*gamma(k2));  @ central chi-square @
else;
  sq=sqrt(q);
  sd=sqrt(psi);
  n=floor(k2);
  a=k2-n;
  i=mbesselei(sq*sd,n,a); @ exponentially scaled Bessel function @
  i=i[.,n];
  f=i.*exp(-((sq-sd).^2)/2).*((sq/sd).^(k2-1))/2; @ non-central chi-square @
endif;
retp(f);
endp;
