/*  BOUND3.PRG 

Gauss Code for "Shrinkage Efficiency Bounds" 
by Bruce E. Hansen
Department of Economics, University of Wisconsin

This program calculates MinRegret for Table 2
*/


/*  BOUND3.PRG */
/* MinRegret */

library pgraph,co;
pqgwin many;
coset;trap 4;

graphset;
_pdate="";
_plwidth=8;
let _pmcolor = 0 0 0 0 0 0 0 0 15;
fonts("complex simgrma");
_pcolor=1;

declare _p,_psi,knots,knot1,_r,m0,m1,ms,m_1;

qn=20000;
pn=20;
psimax=40;
gd=psimax*100;

mregret=zeros(1,pn);

for i (3,pn,1);
  p=i;
  psimaxp=psimax*p;
  psi=seqa(0,psimaxp/gd,gd);
  r2=zeros(gd,1);
  r2[1]=Rbound(p,0,qn);

  for gi (2,gd,1);
    psig=psi[gi];
    r2[gi]=Rbound(p,psig,qn);
  endfor;

  _co_MaxIters=500;
  _co_FeasibleTest=0;
  _co_TrustRadius=.05;
  _co_Options={ BFGS BRENDT FORWARD NONE };
  _co_IneqProc=&RiskBound;  
  knots=seqa(p/10,p/10,20)|seqa(p*2+p/3,p/3,3)|seqa(4*p,p,27);
  n=rows(knots);
  knot1=0|knots[1:n-1]; 
    _p=p;
    _r=r2;
    m_1= TrimmedMomentChiNC(-1,knot1',knots',p,psi);
    m0 = TrimmedMomentChiNC(0,knot1',knots',p,psi);
    m1 = TrimmedMomentChiNC(1,knot1',knots',p,psi);
    ms = ITrimmedMomentChiNC(maxc(knots),p,psi);
    theta0=zeros(n,1);
    theta0[1]= 1;
    theta0[2]= -1;
    {theta0,mr,gr,ret}=co(&MaxRegret6,theta0);
    {theta,mr,gr,ret}=co(&MaxRegret,theta0);
    for it (1,50,1);
      {theta0,mr0,gr,ret}=co(&MaxRegret,theta+rndn(n,1)*.2);
      if mr0<mr;
        mr=mr0;
        theta=theta0;
      endif;
    endfor;
    mregret[i]=mr;

    r=Risk(theta)-r2;
    tit="" $+ "MinRegret Regret, p = " $+ ftocv(p,1,0);
    title(tit);
    xy(psi,r);

    kn=maxc(knots);
    q=seqa(0,kn/99,100);
    qq=q.*(q.<kn)+(q.>=kn);
    x=(qq-knot1').*(qq.>knot1');
    phi=x*theta;
    tit="" $+ "MinRegret Shrinkage Function, p = " $+ ftocv(p,1,0);
    title(tit);
    xy(q,phi);

  format /mb1 /rds 6,3;
  output file=bound3.out reset;
  seqa(3,1,i-2)';
  mregret[.,3:i];
  output off;
 "";


endfor;


/************************************************************/

/************************
Proc MomentChiNC(a,p,psi)
Proc TrimmedMomentChiNC(a,c1,c2,p,psi)
Proc ITrimmedMomentChiNC(a,c,p,psi)

Computes moments, trimmed moments, and CDF of the non-central chi-square distribution

For Q distributed as non-central chi-square with p degrees of freedom
and non-centrality parameter psi

MomentChiNC(a,p,psi)=E[(Q.^a)]
TrimmedMomentChiNC(a,c1,c2,p,psi)=E[(Q.^a)1(c1<=Q<=c2)]
ITrimmedMomentChiNC(c,p,psi)=E[(Q.^-1)1(c<=Q)]

Inputs:
a	moment
c,c1,c2	threshold
p	degrees of freedom
psi	non-centrality parameter

Inputs can be vector or matrix-valued, if element-by-element conformable

c1=0 is okay in TrimmedMomentChiNC(a,c1,c2,p,psi)
Note
TrimmedMomentChiNC(0,0,c,p,psi)=P(Q<=c)

The moments are computed using a truncated series expansion.
The series is computed up to 1000 terms, or until the additional
term is less than a tolerance, set at 1e-8. 

*******************/

proc MomentChiNC(a,p,psi);
local jn,tol,r,f,j,m,f0,r0;
  jn=1000;
  tol=1e-8;
  r0=exp(-psi/2);
  f0=r0;
  r=(2.^a).*gamma(a+p/2)./gamma(p/2).*r0;
  f=r;
  for j (1,jn,1);
    r=r.*psi.*((a+p/2+j-1)./(p/2+j-1)./j./2);
    f=f+r;
    r0=r0.*psi./(j*2);
    f0=f0+r0;
    m=maxc(vec(abs(f0-1)));
    if m<=tol; break;endif;
  endfor;
retp(f);
endp;

proc TrimmedMomentChiNC(a,c1,c2,p,psi);
local jn,tol,r,f,j,d,m,g,cg1,cg2,r0,f0;
  jn=1000;
  tol=1e-8;
  r0=exp(-psi/2);
  f0=r0;
  r=(2.^a).*gamma(a+p/2)./gamma(p/2).*r0;
  cg1=cdfgam(a+p/2,c1/2);
  cg2=cdfgam(a+p/2,c2/2);
  f=r.*(cg2-cg1);
  for j (1,jn,1);
    r=r.*psi.*((a+p/2+j-1)./(p/2+j-1)./j./2);
    cg1=cdfgam(a+p/2+j,c1/2);
    cg2=cdfgam(a+p/2+j,c2/2);
    d=r.*(cg2-cg1);
    f=f+d;
    r0=r0.*psi./(j*2);
    f0=f0+r0;
    m=maxc(vec(abs(f0-1)));
    if m<=tol; break;endif;
  endfor;
retp(f);
endp;


proc ITrimmedMomentChiNC(c,p,psi);
local jn,tol,r,f,j,d,m,g,cg1,r0,f0;
  jn=1000;
  tol=1e-8;
  r0=exp(-psi/2);
  f0=r0;
  r=r0./(p-2);
  cg1=cdfgam(p/2-1,c/2);
  f=r.*(1-cg1);
  for j (1,jn,1);
    r=r.*psi.*((p/2+j-2)./(p/2+j-1)./j./2);
    cg1=cdfgam(p/2+j-1,c/2);
    d=r.*(1-cg1);
    f=f+d;
    r0=r0.*psi./(j*2);
    f0=f0+r0;
    m=maxc(vec(abs(f0-1)));
    if m<=tol; break;endif;
  endfor;
retp(f);
endp;


proc pdfchic(q,r); @ central chi-square density @
local k2,q2,f;
  k2=r/2;
  q2=q/2;
  f=exp(-q2).*(q2.^(k2-1))./(2*gamma(k2)); 
retp(f);
endp;

proc pdfchinc(q,r,psi); @ non-central chi-square density @
local q2,sd,sq,k2,n,a,i,f;
  k2=r/2;
  sq=sqrt(q);
  sd=sqrt(psi);
  n=floor(k2);
  a=k2-n;
  i=mbesselei(sq.*.sd,n,a); @ exponentially scaled Bessel function @
  i=reshape(i[.,n],rows(q),rows(psi));
  f=i.*exp(-((sq-sd').^2)/2).*((sq./sd').^(k2-1))/2; 
retp(f);
endp;

proc Rbound(p,psi,qn);
local q1,q2,ret1,ret2,inc,q,h,f,d,r;
    __output=0;
    __Tol=1e-12;
    _p=p;
    _psi=psi;
    if psi<=p;
      q1=0;ret1=1;
    else;
      q1=(psi-p)*p*(p+2)/(psi^2);
      {q1,ret1}=eqSolve(&hstar1,q1);
    endif;
    q2=2*(p-2)+psi/2-.5*sqrt((psi+4*(p-2))^2-16*((p-2)^2));
    {q2,ret2}=eqSolve(&hstar2,q2);
    inc=(q2-q1)/(qn-1);
    q=seqa(q1,inc,qn);
    h=hstar(q,p,psi);
    f=pdfchinc(q,p,psi); 
    d=sumc((h.^2).*f.*q)*inc;
    r=p-d-8*(p-2)*f[qn];
retp(r);
endp;


proc hstar(q,p,psi);
local h1,q0,sq,sd,n,i,p2;
if psi==0;
  h1=ones(rows(q),1);
else;
  q0=(q.==0);
  sq=sqrt(q+q0);
  sd=sqrt(psi);
  p2=p/2+1;
  n=floor(p2);
  i=mbesselei(sq.*sd,n,p2-n);
  h1=1-(i[.,n]./i[.,n-1]).*(sd./sq);
  h1=h1.*(1-q0)+q0*(1-psi/p);
endif;
retp(h1);
endp;

fn hstar1(q)=hstar(q,_p,_psi);
fn hstar2(q)=hstar(q,_p,_psi)-2*(_p-2)./q;

proc pdfchinc(q,r,psi);
local q2,sd,sq,k2,n,a,i,f;
k2=r/2;
if psi==0;
  q2=q/2;
  f=exp(-q2).*(q2.^(k2-1));
  f=f./(2*gamma(k2));  @ central chi-square @
else;
  sq=sqrt(q);
  sd=sqrt(psi);
  n=floor(k2);
  a=k2-n;
  i=mbesselei(sq*sd,n,a); @ exponentially scaled Bessel function @
  i=i[.,n];
  f=i.*exp(-((sq-sd).^2)/2).*((sq/sd).^(k2-1))/2; @ non-central chi-square @
endif;
retp(f);
endp;


fn MaxRegret(theta)=maxc(Risk(theta)- _r);
fn MaxRegret6(theta)=meanc((Risk(theta)- _r).^6);

proc Risk(theta);
local n,p2,a,b,c,r;
  p2=2*(p-2);
  n=rows(knots);
  a=-cumsumc(theta.*knot1);
  b=cumsumc(theta);
  c=theta'(maxc(knots)-knot1);
  r=p+m_1*(a.*(a-p2))+m0*(2*b.*(a-p))+m1*(b.^2)-ms*c*(p2-c);
retp(r);
endp;

proc RiskBound(theta);
local a,b,f1,f2,f3,f4,f;
  a=-cumsumc(theta.*knot1);
  b=cumsumc(theta);
  f1=2*(p-2)*a-a.^2+(b*4+b*2*(p-2)-2*a.*b).*knots-(b.^2).*(knots.^2);
  f2=2*(p-2)*a-a.^2+(b*4+b*2*(p-2)-2*a.*b).*knot1-(b.^2).*(knot1.^2);
  f3=2*(p-2)-a-b.*knots;
  f4=2*(p-2)-a-b.*knot1;
  f=f1|f2|f3|f4|b[1]|(1-b[1]);
retp(f);
endp;

