/*  BOUND5.PRG 

Gauss Code for "Shrinkage Efficiency Bounds" 
by Bruce E. Hansen
Department of Economics, University of Wisconsin

This program draws the figures (1-3)
It loads saved results created by other programs.

The code as written draws the figures in color. 
To draw in black&white use the alternative _pcolor definitions.
*/

library pgraph;
pqgwin many;
declare _p,_psi;

load aS,abTL2;
abtl=abTL2;
qn=20000;

i=3;
p=i;
gd=121;
psi1=seqa(0,12/(gd-1),gd);
psi=psi1;
r1=zeros(gd,1);
  r1[1]=Rbound(p,0,qn);
  for gi (2,gd,1);
    psig=psi[gi];
    r1[gi]=Rbound(p,psig,qn);
  endfor;

  q2=200; cd=0;
  do until cd .>= .9999;
    q2=q2+100;
    cd=cdfchinc(q2,p+2,sqrt(maxc(psi)));
  endo; 
  inc=q2/(qn-1);
  q=seqa(inc,inc,qn);
  f1=zeros(qn,gd);
  f2=zeros(qn,gd);
  f1[.,1]=pdfchic(q,p)*inc;
  for gi (2,gd,1);
    f1[.,gi]=pdfchinc(q,p,psi[gi])*inc;
    f2[.,gi]=pdfchinc(q,p+2,psi[gi])*(psi[gi]*2*inc);
  endfor;
RiskJS0=Risk_JS0(p,psi);
phi=phi_S(p,q,2);
RiskK=p+f1'(((phi./q)-2).*phi)+f2'(phi./q);
RiskJS=Risk_JS(p,psi,p-2);
RiskTL=Risk_TL(p,psi,abTL[i,1],abTL[i,2]);
risk1=riskJS0~riskK~riskJS~RiskTL;
regret1=risk1-r1;
"p=3: R(B)<R(PL) for psi> " psi1[minindc(risk1[.,3].>risk1[.,4])];


graphset;
_pdate="";
_plwidth=15;
let _pmcolor = 0 0 0 0 0 0 0 0 15;
fonts("complex simgrma");
xlabel("\202y\201");
@let _pcolor = 0 0 0 0 0;@
let _pcolor = 9 2 3 12 13;
let _pltype = 1 2 4 3 6;
let _plegctl = 2 6 6.45 .86;

p=3;
fp=ftocv(p,1,0);
_plegstr="" $+
"R]" $+ fp $+ "[(\202y\201,\202d\201[JS])\000
R]" $+ fp $+ "[(\202y\201,\202d\201[K])\000
R]" $+ fp $+ "[(\202y\201,\202d\201[B])\000
R]" $+ fp $+ "[(\202y\201,\202d\201[TL])\000
R]" $+ fp $+ "[(\202y\201)";
ylabel("Risk");
ytics(0,p,1,10);
xtics(0,4*p,1,10);
xy(psi1,risk1~r1);

i=6;
p=i;
psi2=psi1*2;
psi=psi2;
r2=zeros(gd,1);
  r2[1]=Rbound(p,0,qn);
  for gi (2,gd,1);
    psig=psi[gi];
    r2[gi]=Rbound(p,psig,qn);
  endfor;

  q2=200; cd=0;
  do until cd .>= .9999;
    q2=q2+100;
    cd=cdfchinc(q2,p+2,sqrt(maxc(psi)));
  endo; 
  inc=q2/(qn-1);
  q=seqa(inc,inc,qn);
  f1=zeros(qn,gd);
  f2=zeros(qn,gd);
  f1[.,1]=pdfchic(q,p)*inc;
  for gi (2,gd,1);
    f1[.,gi]=pdfchinc(q,p,psi[gi])*inc;
    f2[.,gi]=pdfchinc(q,p+2,psi[gi])*(psi[gi]*2*inc);
  endfor;
RiskJS0=Risk_JS0(p,psi);
phi=phi_S(p,q,2);
RiskK=p+f1'(((phi./q)-2).*phi)+f2'(phi./q);
RiskJS=Risk_JS(p,psi,p-2);
RiskTL=Risk_TL(p,psi,abTL[i,1],abTL[i,2]);
risk2=riskJS0~riskK~riskJS~RiskTL;
regret2=risk2-r2;
"p=6: R(B)<R(PL) for psi> " psi2[minindc(risk1[.,3].>risk1[.,4])];

p=6;
fp=ftocv(p,1,0);
_plegstr="" $+
"R]" $+ fp $+ "[(\202y\201,\202d\201[JS])\000
R]" $+ fp $+ "[(\202y\201,\202d\201[K])\000
R]" $+ fp $+ "[(\202y\201,\202d\201[B])\000
R]" $+ fp $+ "[(\202y\201,\202d\201[TL])\000
R]" $+ fp $+ "[(\202y\201)";
ytics(0,p,1,10);
xtics(0,4*p,2,10);
xy(psi2,risk2~r2);

graphset;
_pdate="";
_plwidth=15;
let _pmcolor = 0 0 0 0 0 0 0 0 15;
fonts("complex simgrma");
xlabel("\202y\201");
@let _pcolor = 0 0 0 0 ;@
let _pcolor = 9 2 3 12;
let _pltype = 1 2 4 3;
let _plegctl=2 6 5.79 5;

p=3;
fp=ftocv(p,1,0);
_plegstr="" $+
"Regret]" $+ fp $+ "[(\202y\201,\202d\201[JS])\000
Regret]" $+ fp $+ "[(\202y\201,\202d\201[K])\000
Regret]" $+ fp $+ "[(\202y\201,\202d\201[B])\000
Regret]" $+ fp $+ "[(\202y\201,\202d\201[TL])";
ylabel("Regret");
xtics(0,4*p,1,10);
xy(psi1,regret1);

p=6;
fp=ftocv(p,1,0);
_plegstr="" $+
"Regret]" $+ fp $+ "[(\202y\201,\202d\201[JS])\000
Regret]" $+ fp $+ "[(\202y\201,\202d\201[K])\000
Regret]" $+ fp $+ "[(\202y\201,\202d\201[B])\000
Regret]" $+ fp $+ "[(\202y\201,\202d\201[TL])";
yl= "" $+ "Regret";
xtics(0,4*p,2,10);
xy(psi2,regret2);


graphset;
_pdate="";
_plwidth=15;
let _pmcolor = 0 0 0 0 0 0 0 0 15;
fonts("complex simgrma");
@let _pcolor = 0 0 0 0;@
let _pltype = 2 4 1 3;
let _pcolor = 2 3 9 12;

ylabel("Shrinkage Functions");
let _plegctl=2 6 6.8 .47;

p=3;
qn=1001;
qmax=4*p;
inc=qmax/(qn-1);
qq=seqa(0,inc,qn);
phiJS=phi_JS(qq,p);
phiK=phi_S(p,qq,2);
phiS=phi_S(p,qq,aS[p]);
phiTL=phi_TL(qq,abTL[p,.]',p);
phi1=phiK~phiJS~phiS~phiTL;

fp=ftocv(p,1,0);
_plegstr="" $+
"\202f\201[K](q)\000\202f\201[B](q)\000\202f\201[S](q)\000\202f\201[TL](q)";
ytics(0,2,1,10);
xtics(0,qmax,1,10);
xy(qq,phi1);

p=6;
qn=1001;
qmax=4*p;
inc=qmax/(qn-1);
qq=seqa(0,inc,qn);
phiJS=phi_JS(qq,p);
phiK=phi_S(p,qq,2);
phiS=phi_S(p,qq,aS[p]);
phiTL=phi_TL(qq,abTL[p,.]',p);
phi2=phiK~phiJS~phiS~phiTL;

fp=ftocv(p,1,0);
ytics(0,8,1,1);
xtics(0,qmax,2,10);
xy(qq,phi2);


/************************************************************/

fn Risk_JS0(p,psi)=p-((p-2)^2).*MomentChiNC(-1,p,psi);

proc Risk_JS(p,psi,c);
local m0,m1,m2,r;
  m1=TrimmedMomentChiNC(1,0,c,p,psi);
  m0=TrimmedMomentChiNC(0,0,c,p,psi);
  m2=ITrimmedMomentChiNC(c,p,psi);
  r=p+m1-2*p.*m0+c.*(c-2*p+4).*m2;
retp(r);
endp;

proc Risk_TL(p,psi,a,b);
local k1,k2,r,c;
  c=2*(p-2);
  if b>0 and b<1 and a>=0 and a<=c;
    k1=a./(1-b);
    k2=(c-a)./b;
    if k1<k2;
      r=p-2*p*TrimmedMomentChiNC(0,0,k1,p,psi)
         +TrimmedMomentChiNC(1,0,k1,p,psi)
         +a.*(a-c).*TrimmedMomentChiNC(-1,k1,k2,p,psi)
         +2.*b.*(a-p).*TrimmedMomentChiNC(0,k1,k2,p,psi)
         +(b.^2).*TrimmedMomentChiNC(1,k1,k2,p,psi);
    else;
      r=Risk_JS(p,psi,c);
    endif;
  else;
    r=Risk_JS(p,psi,c);
  endif;
retp(r);
endp;

/************************
Proc MomentChiNC(a,p,psi)
Proc TrimmedMomentChiNC(a,c1,c2,p,psi)
Proc ITrimmedMomentChiNC(a,c,p,psi)

Computes moments, trimmed moments, and CDF of the non-central chi-square distribution

For Q distributed as non-central chi-square with p degrees of freedom
and non-centrality parameter psi

MomentChiNC(a,p,psi)=E[(Q.^a)]
TrimmedMomentChiNC(a,c1,c2,p,psi)=E[(Q.^a)1(c1<=Q<=c2)]
ITrimmedMomentChiNC(c,p,psi)=E[(Q.^-1)1(c<=Q)]

Inputs:
a	moment
c,c1,c2	threshold
p	degrees of freedom
psi	non-centrality parameter

Inputs can be vector or matrix-valued, if element-by-element conformable

c1=0 is okay in TrimmedMomentChiNC(a,c1,c2,p,psi)
Note
TrimmedMomentChiNC(0,0,c,p,psi)=P(Q<=c)

The moments are computed using a truncated series expansion.
The series is computed up to 1000 terms, or until the additional
term is less than a tolerance, set at 1e-8. 

*******************/

proc MomentChiNC(a,p,psi);
local jn,tol,r,f,j,m,f0,r0;
  jn=1000;
  tol=1e-8;
  r0=exp(-psi/2);
  f0=r0;
  r=(2.^a).*gamma(a+p/2)./gamma(p/2).*r0;
  f=r;
  for j (1,jn,1);
    r=r.*psi.*((a+p/2+j-1)./(p/2+j-1)./j./2);
    f=f+r;
    r0=r0.*psi./(j*2);
    f0=f0+r0;
    m=maxc(vec(abs(f0-1)));
    if m<=tol; break;endif;
  endfor;
retp(f);
endp;

proc TrimmedMomentChiNC(a,c1,c2,p,psi);
local jn,tol,r,f,j,d,m,g,cg1,cg2,r0,f0;
  jn=1000;
  tol=1e-8;
  r0=exp(-psi/2);
  f0=r0;
  r=(2.^a).*gamma(a+p/2)./gamma(p/2).*r0;
  cg1=cdfgam(a+p/2,c1/2);
  cg2=cdfgam(a+p/2,c2/2);
  f=r.*(cg2-cg1);
  for j (1,jn,1);
    r=r.*psi.*((a+p/2+j-1)./(p/2+j-1)./j./2);
    cg1=cdfgam(a+p/2+j,c1/2);
    cg2=cdfgam(a+p/2+j,c2/2);
    d=r.*(cg2-cg1);
    f=f+d;
    r0=r0.*psi./(j*2);
    f0=f0+r0;
    m=maxc(vec(abs(f0-1)));
    if m<=tol; break;endif;
  endfor;
retp(f);
endp;


proc ITrimmedMomentChiNC(c,p,psi);
local jn,tol,r,f,j,d,m,g,cg1,r0,f0;
  jn=1000;
  tol=1e-8;
  r0=exp(-psi/2);
  f0=r0;
  r=r0./(p-2);
  cg1=cdfgam(p/2-1,c/2);
  f=r.*(1-cg1);
  for j (1,jn,1);
    r=r.*psi.*((p/2+j-2)./(p/2+j-1)./j./2);
    cg1=cdfgam(p/2+j-1,c/2);
    d=r.*(1-cg1);
    f=f+d;
    r0=r0.*psi./(j*2);
    f0=f0+r0;
    m=maxc(vec(abs(f0-1)));
    if m<=tol; break;endif;
  endfor;
retp(f);
endp;


/***********
Proc igam(a,x)

igam(a,x)=gamma(a,x)*exp(x)/x^a

where gamma(a,x) is the incomplete gamma function

*******************/

proc igam(a,x);
local tol,d,s2,ii,s,r,j;
  ii=1000;
  tol=1e-12;
  r=1./a;
  s=r;
  for j (1,ii,1);
    r=r.*(x./(j+a));
    s=s+r;
    d=maxc(vec(abs(r)));
    if d<=tol; break; endif;
  endfor;
retp(s);
endp;

fn phi_S(p,q,a)=(p-2*a+2)-2./igam((p/2-a+1),q/2);

proc pdfchic(q,r); @ central chi-square density @
local k2,q2,f;
  k2=r/2;
  q2=q/2;
  f=exp(-q2).*(q2.^(k2-1))./(2*gamma(k2)); 
retp(f);
endp;

proc pdfchinc(q,r,psi);	@ non-central chi-square density @
local q2,sd,sq,k2,n,a,i,f;
  k2=r/2;
  sq=sqrt(q);
  sd=sqrt(psi);
  n=floor(k2);
  a=k2-n;
  i=mbesselei(sq.*.sd,n,a); @ exponentially scaled Bessel function @
  i=reshape(i[.,n],rows(q),rows(psi));
  f=i.*exp(-((sq-sd').^2)/2).*((sq./sd').^(k2-1))/2; 
retp(f);
endp;

fn phi_JS(q,p)=(p-2).*(q.>p-2)+q.*(q.<=p-2);

proc phi_TL(q,theta,p);
local a,b,c,phi;
  a=theta[1];b=theta[2];
  c=2*(p-2);
  phi=a+q*b;
  phi=phi.*(phi.<=q)+q.*(phi.>q);
  phi=phi.*(phi.<=c)+c.*(phi.>c);
retp(phi);
endp;

proc Rbound(p,psi,qn);
local q1,q2,ret1,ret2,inc,q,h,f,d,r;
    __output=0;
    __Tol=1e-12;
    _p=p;
    _psi=psi;
    if psi<=p;
      q1=0;ret1=1;
    else;
      q1=(psi-p)*p*(p+2)/(psi^2);
      {q1,ret1}=eqSolve(&hstar1,q1);
    endif;
    q2=2*(p-2)+psi/2-.5*sqrt((psi+4*(p-2))^2-16*((p-2)^2));
    {q2,ret2}=eqSolve(&hstar2,q2);
    inc=(q2-q1)/(qn-1);
    q=seqa(q1,inc,qn);
    h=hstar(q,p,psi);
    f=pdfchinc(q,p,psi); 
    d=sumc((h.^2).*f.*q)*inc;
    r=p-d-8*(p-2)*f[qn];
retp(r);
endp;


proc hstar(q,p,psi);
local h1,q0,sq,sd,n,i,p2;
if psi==0;
  h1=ones(rows(q),1);
else;
  q0=(q.==0);
  sq=sqrt(q+q0);
  sd=sqrt(psi);
  p2=p/2+1;
  n=floor(p2);
  i=mbesselei(sq.*sd,n,p2-n);
  h1=1-(i[.,n]./i[.,n-1]).*(sd./sq);
  h1=h1.*(1-q0)+q0*(1-psi/p);
endif;
retp(h1);
endp;

fn hstar1(q)=hstar(q,_p,_psi);
fn hstar2(q)=hstar(q,_p,_psi)-2*(_p-2)./q;

